#include "trainserver.h"
#include <zeitgeist/logserver/logserver.h>
#include <zeitgeist/corecontext.h>
#include <oxygen/simulationserver/simcontrolnode.h>
#include "trainsystem.h"

using namespace oxygen;
using namespace zeitgeist;
using namespace std;
using namespace boost;
using namespace salt;

TrainServer::TrainServer()
{
}

TrainServer::~TrainServer()
{
}

void TrainServer::OnLink()
{
    mSimulationServer = shared_dynamic_cast<SimulationServer>
        (GetCore()->Get(gServerPath + "simulation"));
    if (mSimulationServer.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) ERROR: SimulationServer not found\n";
    }

    SetupRobotPartMap();
    
    mActiveId = "1";
}

void TrainServer::OnUnlink()
{
    mTrainSystem.reset();
}

bool TrainServer::Init(const string& trainSysName)
{
    GetLog()->Normal() << "(TrainServer) Init " << trainSysName << "\n";
    mTrainSystem.reset();

    if (! trainSysName.empty())
    {
        // create the OpenGLSystem
        mTrainSystem = shared_dynamic_cast<TrainSystem>
            //(GetCore()->New(trainSysName));
            (GetCore()->CreateContext()->
                New(trainSysName, gServerPath + "trainserver/" + trainSysName)
            );

        if (mTrainSystem.get() == 0)
        {
            // could not create OpenGLSystem
            GetLog()->Error() << "(TrainServer) ERROR: unable to create "
                              << trainSysName << "\n";
            return false;
        }

        if (mTrainSystem->Init() == false)
        {
            GetLog()->Error() << "(TrainServer) ERROR: unable to initialize "
                              << trainSysName << "\n";
            return false;
        }
    }

    return true;
}

void TrainServer::Update(){
    if (mTrainSystem.get() == 0)
    {
        // GetLog()->Error() << "(TrainServer) ERROR: unable to update "
        //     << "no TrainSystem registered" << endl;
        return ;
    }
    
    const float maxTrainingSeconds =  60.0f * 20.0f; // reset after this time (20 mins)
	if (mSimulationServer.get() != 0){
        if (mSimulationServer->GetTime() > maxTrainingSeconds){
            ResetTime();
            mTrainSystem->Reset();
        }else mTrainSystem->Update();
	}

}

void TrainServer::ResetTime(){
    if (mSimulationServer.get() != 0){
        mSimulationServer->ResetTime();
        for ( TLeafList::iterator iter=mSimulationServer->begin(); 
              iter != mSimulationServer->end(); ++iter ){
            shared_ptr<SimControlNode> ctrNode =  shared_dynamic_cast<SimControlNode>(*iter);
            if (ctrNode.get() == 0) continue;

            ctrNode->SetTime(0.0);
        }
    }
}

void TrainServer::Quit(){
    if (mSimulationServer.get() != 0)
     mSimulationServer->Quit();
}

void TrainServer::SetupRobotPartMap(){
    mRobotPartMap.clear();
    mRobotPartMap[PART_HEAD] 		= "head";
    mRobotPartMap[PART_NECK]        = "neck";
    mRobotPartMap[PART_BODY] 		= "body";
    mRobotPartMap[PART_LSHOULDER] 	= "lshoulder";
    mRobotPartMap[PART_RSHOULDER] 	= "rshoulder";
    mRobotPartMap[PART_LUPPERARM] 	= "lupperarm";
    mRobotPartMap[PART_RUPPERARM] 	= "rupperarm";
    mRobotPartMap[PART_LELBOW] 	    = "lupperarm";
    mRobotPartMap[PART_RELBOW]      = "rupperarm";
    mRobotPartMap[PART_LLOWERARM] 	= "llowerarm";
    mRobotPartMap[PART_RLOWERARM] 	= "rlowerarm";
    mRobotPartMap[PART_LHIP_1] 		= "lhip1";
    mRobotPartMap[PART_RHIP_1] 		= "rhip1";
    mRobotPartMap[PART_LHIP_2] 		= "lhip2";
    mRobotPartMap[PART_RHIP_2] 		= "rhip2";
    mRobotPartMap[PART_LTHIGH] 		= "lthigh";
    mRobotPartMap[PART_RTHIGH] 		= "rthigh";
    mRobotPartMap[PART_LSHANK] 		= "lshank";
    mRobotPartMap[PART_RSHANK] 		= "rshank";
    mRobotPartMap[PART_LANKLE] 		= "lankle";
    mRobotPartMap[PART_RANKLE] 		= "rankle";
    mRobotPartMap[PART_LFOOT] 		= "lfoot";
    mRobotPartMap[PART_RFOOT] 		= "rfoot";
}

void TrainServer::SetActiveId(const string& id){
    if (! id.empty())
     mActiveId = id;
}

shared_ptr<Transform> TrainServer::GetRobotPart(RobotPart part) const{
    TRobotPartMap::const_iterator idIter = mRobotPartMap.find(part);
    if (idIter == mRobotPartMap.end())
    {
        GetLog()->Error()
            << "(TrainServer) unknown robot part!" << endl;
        return shared_ptr<Transform>();
    }

    string name = (*idIter).second;

    stringstream ss;
    ss << gScenePath << "AgentAspect" << mActiveId << "/spacenao/" << name;
    shared_ptr<Transform> transform = shared_dynamic_cast<Transform>
        (GetCore()->Get(ss.str()));
    if (transform.get() == 0)
    {
        return shared_ptr<Transform>();
    }

    return transform;
}

shared_ptr<Body> TrainServer::GetRobotPartBody(RobotPart part) const
{
    shared_ptr<Transform> transform = GetRobotPart(part);
    if (transform.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) RobotPart "<<part<<" Transform get failed!\n";
        return shared_ptr<Body>();
    }

    shared_ptr<Body> body = shared_dynamic_cast<Body>
        (transform->GetChildOfClass("Body"));
    if (body.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) RobotPart "<<part<<" Body get failed!\n";
        return shared_ptr<Body>();
    }

    return body;
}

shared_ptr<Joint> TrainServer::GetRobotPartJoint(RobotPart part) const
{
    shared_ptr<Transform> transform = GetRobotPart(part);
    if (transform.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) RobotPart Transform get failed!\n";
        return shared_ptr<Joint>();
    }

    shared_ptr<Joint> joint = shared_dynamic_cast<Joint>
        (transform->GetChildSupportingClass("Joint"));
    if (joint.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) RobotPart Joint get failed!\n";
        return shared_ptr<Joint>();
    }

    return joint;
}

Matrix TrainServer::GetRobotPartMatrix(RobotPart part) const
{
    Matrix m;
    shared_ptr<Body> body = GetRobotPartBody(part);
    if (body.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) Body get failed!\n";
        return m;
    }

    m = body->GetRotation();
    m.Pos() = body->GetPosition();
    return m;
}

Vector3f TrainServer::GetRobotPartPos(RobotPart part) const
{
    shared_ptr<Body> body = GetRobotPartBody(part);
    if (body.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) Body get failed!\n";
        return Vector3f(0.0f, 0.0f, 0.0f);
    }
    return body->GetPosition();
    
}

shared_ptr<Transform> TrainServer::GetBall() const
{
    shared_ptr<Transform> transform = shared_dynamic_cast<Transform>
        (GetCore()->Get(gScenePath + "Ball"));
    if (transform.get() == 0)
    {
        GetLog()->Error()
            << "(TrainServer) Ball Transform get failed!\n";
        return shared_ptr<Transform>();
    }

    return transform;
}

shared_ptr<Body> TrainServer::GetBallBody() const{
    shared_ptr<Transform> transform = GetBall();
    if (transform.get() == 0)
      return shared_ptr<Body>();

    shared_ptr<Body> body = shared_dynamic_cast<Body>
        (transform->GetChildOfClass("Body"));
    if (body.get() == 0) 
        GetLog()->Error() << "(TrainServer) Ball Body get failed!\n";
    return body;
}

Vector3f TrainServer::GetBallPos() const{
    shared_ptr<Body> body = GetBallBody();
    if (body.get() == 0)
      return Vector3f(-1000.0, -1000.0, -1000.0);


    return body->GetPosition();
}

bool TrainServer::MoveAndRotateAgent(const Vector3f& pos, float angle){
    Vector3f agentPos = GetRobotPartPos(PART_BODY);
    shared_ptr<Transform> parent = shared_dynamic_cast<Transform>
        (GetRobotPart(PART_BODY)->FindParentSupportingClass<Transform>().lock());

    if (parent.get() == 0){
        GetLog()->Error()
            << "(TrainServer) MoveAndRotateAgent ERROR: can't get parent node.\n";
        return false;
    }

    Leaf::TLeafList leafList;

    parent->ListChildrenSupportingClass<Body>(leafList, true);

    if (leafList.size() == 0){
        GetLog()->Error()
            << "(TrainServer) MoveAndRotateAgent ERROR: agent aspect doesn't have "
            << "children of type Body\n";

        return false;
    }

    shared_ptr<Body> body;
    body = GetRobotPartBody(PART_BODY);

    Matrix bodyR = body->GetRotation();
    bodyR.InvertRotationMatrix();
    Matrix mat;
    mat.RotationZ(gDegToRad(angle));
    mat *= bodyR;

    Leaf::TLeafList::iterator iter = leafList.begin();

    // move all child bodies
    for (; iter != leafList.end();  ++iter){
        shared_ptr<Body> childBody =
            shared_dynamic_cast<Body>(*iter);

        Vector3f childPos = childBody->GetPosition();
        Matrix childR = childBody->GetRotation();
        childR = mat*childR;
        childBody->SetPosition(pos + mat.Rotate(childPos-agentPos));
        childBody->SetVelocity(Vector3f(0,0,0));
        childBody->SetAngularVelocity(Vector3f(0,0,0));
        childBody->SetRotation(childR);
    }

    return true;
}

void TrainServer::DropBall(const Vector3f& pos){
    shared_ptr<Body> body = shared_dynamic_cast<Body>
        (GetCore()->Get(gScenePath + "Ball/physics"));

    body->SetPosition(pos);
    body->SetVelocity(Vector3f(0,0,0));
    body->SetAngularVelocity(Vector3f(0,0,0));
}

void TrainServer::SetBallVel(const salt::Vector3f& vel){
    shared_ptr<Body> body = shared_dynamic_cast<Body>
        (GetCore()->Get(gScenePath + "Ball/physics"));

    body->Enable();
    body->SetVelocity(vel);
}

bool TrainServer::SaveRobotScene(){
    /** clear robot scene map */
    mRobotSceneMap.clear();

    shared_ptr<Transform> parent = shared_dynamic_cast<Transform>
        (GetRobotPart(PART_BODY)->FindParentSupportingClass<Transform>().lock());

    if (parent.get() == 0){
        GetLog()->Error() << "(TrainServer) SaveRobotScene ERROR: can't get parent node.\n";
        return false;
    }

    Leaf::TLeafList leafList;

    parent->ListChildrenSupportingClass<Body>(leafList, true);

    if (leafList.size() == 0){
        GetLog()->Error()
            << "(TrainServer) SaveRobotScene ERROR: agent aspect doesn't have "
            << "children of type Body\n";

        return false;
    }

    Leaf::TLeafList::iterator iter = leafList.begin();

    // save all child bodies
    for (;iter != leafList.end();++iter){
        shared_ptr<Body> body = shared_dynamic_cast<Body>(*iter);

        mRobotSceneMap[body] = body->GetRotation();
        mRobotSceneMap[body].Pos() = body->GetPosition();
    }

    return true;
}

bool TrainServer::RestoreRobotScene(){
    shared_ptr<Transform> parent = shared_dynamic_cast<Transform>
        (GetRobotPart(PART_BODY)->FindParentSupportingClass<Transform>().lock());

    if (parent.get() == 0){
        GetLog()->Error()
            << "(TrainServer) RestoreRobotScene ERROR: can't get parent node.\n";
        return false;
    }

    Leaf::TLeafList leafList;

    parent->ListChildrenSupportingClass<Body>(leafList, true);

    if (leafList.size() == 0){
        GetLog()->Error()
            << "(TrainServer) RestoreRobotScene ERROR: agent aspect "
            << "doesn't have children of type Body\n";

        return false;
    }

    Leaf::TLeafList::iterator iter = leafList.begin();

    // restore all child bodies
    for ( ; iter != leafList.end(); ++iter){
        shared_ptr<Body> body = shared_dynamic_cast<Body>(*iter);

        TRobotSceneMap::iterator body_iter = mRobotSceneMap.find(body);
        if (body_iter == mRobotSceneMap.end()){
            GetLog()->Error()
                << "(TrainServer) RestoreRobotScene: unknown robot part!"
                << endl;
            continue;
        }

        const Matrix& m = body_iter->second;
        body->SetPosition(m.Pos());
        body->SetVelocity(Vector3f(0,0,0));
        body->SetAngularVelocity(Vector3f(0,0,0));
        body->SetRotation(m);
    }

    return true;
}

