#ifndef __Q_LEARNING_H__
#define __Q_LEARNING_H__


//Whether the Q-Learner is used in a debug mode
#ifndef ISDEBUG
  #define ISDEBUG 1  
#endif

const bool PSEUDO_RANDOM_TRYNEW = 1;



#define RANDOMSEED 54361


#include <map> 
#include <assert.h>
#include <set>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <float.h>
#include <utils/smp_func.h>

template <typename StateT, typename ActionT, typename QValueT> 
class QTable{
public:
    typedef std::pair<StateT, ActionT>  StateActionPair;
    typedef std::map<StateActionPair, QValueT> Table;
    typedef std::set<ActionT> ActionSet;
    typedef std::set<StateT> StateSet;

    typedef std::map<ActionT, bool> ActionVisitSet; //whether an action is visited
    typedef std::map<StateT,ActionVisitSet> StateActionVisitMap; 
private:
    std::string tag;
    StateSet stateSet;
    ActionSet actionSet;
    QValueT betterQValue(QValueT q1, QValueT q2){ // used to compare two q values; it can be externed if necessary 
        if (q2 < q1) return q1; // the bigger, the better
        //if (q1 < q2) return q1; // the less, the better;
        return q2;
    }

public:
    void setTag(std::string tag){ this->tag = tag;};
    std::string getTag(){ return this->tag;};

    StateActionVisitMap visitMap;  //used for random visit
    StateT  NULLSTATE;
    ActionT NULLACTION;
    QValueT NULLQVALUE;
    Table   table;
    bool    prepared;//whether an action is visited
    bool    readOK;

    void clearAll(){
		table.clear();
		stateSet.clear();
		actionSet.clear();
		visitMap.clear();
		prepared=false;
        readOK=false;
	};

	long getStateActionSize(){return table.size();};
	long getStateSize(){return stateSet.size();};
	long getActionSize(){return actionSet.size();};
	bool ifStateActionPairVisited(StateT state, ActionT action){
		if (!prepared) return false;
		if (!isPairExist(state,action)) return false;
     	return visitMap[state][action];
	}

    void prepareForLearning(){ // make preparation for learning
        if (table.empty() || prepared) return; // do nothing;

        typename StateSet::iterator it_s;
        for (it_s = stateSet.begin(); it_s != stateSet.end(); it_s++ ){
            ActionVisitSet tempActionSet;
            for (typename Table::iterator it_t = table.begin(); it_t != table.end(); it_t++)
                if (*it_s == it_t->first.first)
                    visitMap[*it_s][it_t->first.second] =false;
        }

        prepared = true;
    };

    QTable(){
        NULLSTATE = StateT();
        NULLACTION = ActionT();
        NULLQVALUE = QValueT();
        prepared= false;
        readOK=false;
        tag="General_QTable";
        srand(RANDOMSEED);
    };

    ~QTable(){};

	bool addStateActionPair(StateT state, ActionT action, QValueT value, bool isVisited){ 
        if (!addStateActionPair(state,action,value)) return false;
        visitMap[state][action]=isVisited;
		return true;
	};

    bool addStateActionPair(StateT state, ActionT action, QValueT value){ 
        if (isPairExist(state,action)) return false;
        table[StateActionPair(state,action)]=value;
		if (!isStateExist(state)) stateSet.insert(state);
		if (!isActionExist(action)) actionSet.insert(action);
        return true;
    };
    bool addStateActionPair(StateT state, ActionT action){ 
        return addStateActionPair(state, action, NULLQVALUE);
    };
    bool delStateActionPair(StateT state, ActionT action){
        if (!isPairExist(state,action)) return false;
        table.erase(std::pair<StateT,ActionT>(state,action));
        visitMap[state].erase(action);
        return true;
    };

    bool addState(StateT state){ //when a state is available for all the actions
        if (isStateExist(state)) return false;
        else {
            stateSet.insert(state);
            if (!actionSet.empty()){
                typename ActionSet::iterator it;
                for (it = actionSet.begin(); it != actionSet.end(); it++ )
                    addStateActionPair(state,*it);
            }
        }
        return true;
    };

    bool addAction(ActionT action){//when an action is available for all the states
        if (isActionExist(action)) return false;
        else {
            actionSet.insert(action);
            if (!stateSet.empty()){
                typename StateSet::iterator it;
                for (it = stateSet.begin(); it != stateSet.end(); it++ )
                    addStateActionPair(*it,action);
            }
        }
        return true;
    };

    bool delState(StateT state){
        if (!isStateExist(state)) return false;
        else {
            stateSet.erase(state);
            if (!table.empty()){
                for(typename Table::iterator it = table.begin(); it!=table.end();){
                    if (it->first.first == state) table.erase((it++)->first);
                    else ++it;
                }
            }
        }
        if (prepared) visitMap.erase(state);
        return true;
    };

    bool delAction(ActionT action){
        if (!isActionExist(action)) return false;
        else {
            actionSet.erase(action);
            if (!table.empty()){
				 for(typename Table::iterator it = table.begin(); it!=table.end();){
                    if (it->first.second == action) table.erase((it++)->first);
                     else ++it;
                }
            }
        }
		if (prepared) {
			for (typename StateActionVisitMap::iterator it = visitMap.begin(); it!=visitMap.end(); it++){
				if (it->second.find(action)!=it->second.end()) it->second.erase(action);
			}
		}
        return true;
    }

    bool isPairExist(StateT state, ActionT action){
        typename Table::iterator iter = table.find(std::pair<StateT,ActionT>(state,action));
        return (iter != table.end());
    };
    bool isPairExist(StateT state){
        if (!table.empty()){
              for (typename Table::iterator it = table.begin(); it != table.end(); it++)
                if (state == it->first.first) return true;

        }
        return false;
    };

    ActionT getBestAction(StateT state){ // find the best action
#ifdef ISDEBUG
        assert(isPairExist(state));
#endif
        ActionT bestaction;
        QValueT q_value;
        bool firstrun=true;
          for (typename Table::iterator it = table.begin(); it != table.end(); it++)
            if (state == it->first.first) {
                if (firstrun){
                    bestaction = it->first.second;
                    q_value = it->second;
                    firstrun=false;
                }
                else if (q_value != betterQValue(it->second, q_value)){
                    bestaction = it->first.second;
                    q_value = it->second;
                }

            }
            return bestaction;
    }

    QValueT getBestValue(StateT state){ // find the best action
#ifdef ISDEBUG
        assert(isPairExist(state));
#endif      
        QValueT q_value;
        bool firstrun=true;
          for (typename Table::iterator it = table.begin(); it != table.end(); it++)
            if (state == it->first.first) {
                if (firstrun){
                    q_value = it->second;
                    firstrun=false;
                }
                else if (q_value != betterQValue(it->second, q_value))
                    q_value = it->second;
            }
            return q_value;
    }

    ActionT getRandomAction(StateT state, bool preferTryNew){ // find a random action
#ifdef ISDEBUG
         assert(isPairExist(state));
#endif
         ActionT action;
         typename StateActionVisitMap::iterator it_v= visitMap.find(state);
		 if (it_v == visitMap.end())
		    std::cout<<"[ERROR] QTable Invalid State ( " <<state <<" ) !"<<std::endl;
         assert(it_v != visitMap.end());

         if (PSEUDO_RANDOM_TRYNEW && preferTryNew){
             for (typename ActionVisitSet::iterator it_av= it_v->second.begin(); it_av != it_v->second.end(); it_av++)
                 if (it_av->second == false) {
                     action = it_av->first;
                     break;
                 }
                 if (action == NULLACTION){
                     for (typename ActionVisitSet::iterator it_av= it_v->second.begin(); it_av != it_v->second.end(); it_av++)
                         it_av->second = false;
                     action = it_v->second.begin()->first;
                 }
         }else{
             std::vector<ActionT> tempActionSet;
             for (typename ActionVisitSet::iterator it_av= it_v->second.begin(); it_av != it_v->second.end(); it_av++)
                 if (it_av->second == false) tempActionSet.push_back(it_av->first);

             if(tempActionSet.empty()){
                 for (typename ActionVisitSet::iterator it_av= it_v->second.begin(); it_av != it_v->second.end(); it_av++){
                     it_av->second = false;
                     tempActionSet.push_back(it_av->first);
                 }
             }

             int index = rand()%tempActionSet.size();
             action = tempActionSet[index];
         }
        if (preferTryNew) visitMap[state][action] = true;
        return action;
    }

    bool isStateExist(StateT state){
        typename StateSet::iterator iter = stateSet.find(state);
        return (iter != stateSet.end());
    };
    bool isActionExist(ActionT action){
        typename ActionSet::iterator iter = actionSet.find(action);
        return (iter != actionSet.end());
    };

    QValueT getValue(StateT state, ActionT action){
#ifdef ISDEBUG
        assert(isPairExist(state,action));
#endif
        typename Table::iterator iter = table.find(StateActionPair(state,action));
        return iter->second;
    };

    bool setValue(StateT state, ActionT action, QValueT value){
        if (!isPairExist(state,action)) {
            addStateActionPair(state,action,value);
            return false;
        }
        else {
            table[StateActionPair(state,action)]=value;
            return true;
        }
    };
    float getUnVisitedPairRate(){
        if (getStateActionSize() == 0) return 0;
        int count = 0;
        for (typename Table::iterator it = table.begin(); it != table.end(); it++){
            if (it->second == NULLQVALUE)
                count++;
        }
        return (float)count / (float) getStateActionSize();
    }

    bool isReadSuccess(){return readOK;};
    bool save(const char* filename){
        std::ofstream out;
        out.open(filename);

        if(!out) {
            std::cout << "Cannot open output file : " << filename << std::endl;
            return false;
        }
        out << (*this);
        out.close();
        return true;
    }
    bool read(const char* filename){
        std::ifstream in;
        in.open(filename);

        if(!in) {
            std::cout << "Cannot open input file : " << filename << std::endl;
            return false;
        }
        in >> (*this);
        in.close();
        return isReadSuccess();
    }


    friend std::ostream& operator<<(std::ostream& os, QTable &QT){
        os <<" [ "<<QT.tag<<" ] "<<" Total Pairs:"   << QT.getStateActionSize()
            <<" unVisitedRate:" << QT.getUnVisitedPairRate() <<std::endl;
        for (typename QTable<StateT,ActionT,QValueT>::Table::iterator it = QT.table.begin(); it != QT.table.end(); it++){
            os  <<" S:(" << it->first.first << ")"
                <<" A:(" << it->first.second << ")"
                <<" Q:" << it->second  
                <<" Visited:" << QT.ifStateActionPairVisited(it->first.first, it->first.second) << std::endl;
        }
      return os;
    }

    friend std::istream& operator>>(std::istream& is, QTable &QT){
        long _size = 0;
        float _unVisitedRate = 0.0f;
		bool _isVisited = false;
        StateT s = QT.NULLSTATE;
        ActionT a = QT.NULLACTION;
        QValueT q =  QT.NULLQVALUE;
        is.ignore(50, '[');         is >> QT.tag;
        is.ignore(50, ':');         is >> _size;
        is.ignore(50, ':');         is >> _unVisitedRate;
        if (_size> 0){
            for (int i=0; i<_size && !is.eof(); i++){
                is.ignore(50, '(');     is >> s;
                is.ignore(50, '(');     is >> a;
                is.ignore(50, ':');     is >> q;
                is.ignore(50, ':');     is >> _isVisited;
                QT.addStateActionPair(s,a,q,_isVisited);
            }
            QT.readOK=true;
        }
        if (_size <= 0 || _size!= QT.getStateActionSize() ) {
            QT.readOK = false;
            std::cout<<"[ERROR] Q-Table does not match! Expected size="<<_size<<" Size now="
                     << QT.getStateActionSize()<<std::endl;

        }
        return is;
    };
    
};


template <typename StateT, typename ActionT, typename QValueT> 
class QLearner     
{
protected :
    QTable<StateT,ActionT,QValueT> *Q;
    float alpha; //Holds the alpha parameter: the learning rate
    float gamma; //Holds the gamma parameter: the discount rate
    float rho; //Holds the rho parameter: exploration randomness
    bool preferTryNew;
    float random0_1(){ return float(rand())/RAND_MAX; }

    ActionT getAction(StateT s){
		if (mIsExecuteOnly) return Q->getBestAction(s);

		if (random0_1() < rho){
            return Q->getRandomAction(s, preferTryNew);
		}
		else { 
            return Q->getBestAction(s);
			
		}
    };

    void update( StateT s, ActionT a, StateT result_s, QValueT reward){

        // Get the current q value
        QValueT q = Q->getValue(s, a);

        // Get the q of the best action from the new state
        QValueT maxQ = Q->getBestValue(result_s);

        // recalculate the q
        q = ((QValueT)1.0f - alpha) * q + alpha * (reward + gamma * maxQ);

        // Store the new Q value
        Q->setValue(s, a, q);
    };
public:
    int episode;	
	bool mIsExecuteOnly;
    QLearner(QTable<StateT,ActionT,QValueT> *Q, float alpha, float gamma, float rho, bool preferTryNew){
        this->Q = Q;
        this->alpha = alpha;
        this->gamma = gamma;
        this->rho = rho;
        this->preferTryNew = preferTryNew;

        this->episode = 0;
		this->mIsExecuteOnly = false;
        current_state = init_state = last_state = Q->NULLSTATE;
        current_action = last_action = Q->NULLACTION;
    };

    ~QLearner(){};

    StateT current_state;
    StateT init_state;
    ActionT current_action;
    ActionT last_action;
    StateT last_state;
	void resetState(){
		current_state = init_state;
	};

    void setInitState(StateT initState){
		init_state = initState;
        current_state = initState;
        if (!Q->prepared) Q->prepareForLearning();
		std::cout<<"[INIT] << Total State-Action pairs: " << Q->getStateActionSize() <<std::endl;
		std::cout<<"[INIT] << States (" << Q->getStateSize() <<") "
			          << "Actions (" << Q->getActionSize() <<") "<<std::endl;
		//current_action = getAction();  
    };

    //Whether the Q-Learner prefers to explore the actions that has not been visited
    ActionT getAction(){
        return getAction(current_state);
    };

    ActionT getRandomAction(StateT s){
       return Q->getRandomAction(s, preferTryNew);
    };

    ActionT getBestAction(StateT s){
        return Q->getBestAction(s);
    };

    virtual StateT doAction(const ActionT& action)=0;
    virtual QValueT getReward()=0;
    virtual bool isFinish()=0;
	virtual StateT judgeState()=0;
    virtual bool putAdditionalInfo(std::ofstream& os)=0;
    virtual bool getAdditionalInfo(std::ifstream& is)=0;

    void run(){
        while(!isFinish()){ 
            doIteration();
        }
    }

    void doIteration(){     
//Iteration Called When The Next State will be known immediately S->  A->S'->R  ->...
//                                                                | doIteration |
        last_state = current_state;
	    last_action = current_action;
        current_action = getAction();  
        current_state = doAction(current_action);
		QValueT reward = getReward();
		if (!mIsExecuteOnly) {
			update(last_state, current_action, current_state, reward);
			episode++;
		}
    }

	void doDelayIteration(bool isFirst, bool rewardOnly = false){     
//Iteration Called When The Next State Cannot be known immediately  S-> A-> S'->R -> A->
//                                                                         |doIteration |
		last_action = current_action;
        last_state = current_state;
		current_state = judgeState();
		if (mIsExecuteOnly){
          if (!rewardOnly) current_action  = getAction();
		  std::cout << "ExecuteOnly" <<"["<< this->episode <<"] "
			  <<" S_Now=("<<current_state << ")"
			  <<" ChangeAction ( "<< current_action << " ) " <<std::endl;
		}else {
			if (isFirst){
				if (!rewardOnly) current_action  = getAction();
				std::cout <<"["<< this->episode <<"] "
					<<"The First Iteration! "
					<<" S_Now=("<<current_state << ")"
					<<" Choose Action= ( "<< current_action << " ) "
					<<std::endl;
            }else{  
                    if (last_action==Q->NULLACTION)
                     std::cout<< "[ERROR] QLeaner: INVALID Last_Action ( "<<last_action<<" ); reward cannot be given!"<<std::endl;

                    assert(!(last_action==Q->NULLACTION));
					QValueT reward = getReward();
					std::cout <<"["<< this->episode <<"] "
						<<"S_last=("<<last_state << ")"
						<<" A_Last= ( "<<last_action << " ) "
						<<" S_Now=("<<current_state << ")"
						<<" R="<< reward <<std::endl;
					update(last_state, last_action, current_state, reward);
					if (!rewardOnly) current_action  = getAction();
					episode++;
			}
		}
		if (!rewardOnly) doAction(current_action);
	}

	bool saveTable(const char* filename, bool isAppendMode = false){
        std::ofstream out;
		if (isAppendMode)  
              out.open(filename, std::ios::out|std::ios::app);
        else  out.open(filename);

		if(!out) {
			std::cout << "Cannot open output file : " << filename << std::endl;
			return false;
		}
        if (!putAdditionalInfo(out)){
            std::cout<<"[ERROR] QLeaner: Cannot save Additional Information!"<<std::endl;
        }
        out <<" Episode:"<< episode
            <<" alpha:"<< alpha
            <<" gamma:"<< gamma 
            <<" rho:"<< rho 
			<<" preferTryNew:"<< preferTryNew
            << std::endl
            << *Q;
 		out.close();
		return true;
	}

    bool releaseTable(const char* filename, bool isAppendMode = false){
        typedef typename QTable<StateT,ActionT,QValueT>::StateActionVisitMap::iterator VMIterator;
        typedef typename QTable<StateT,ActionT,QValueT>::ActionVisitSet::iterator AVIterator;
        if(!Q->prepared){
            std::ofstream out(filename);
            out.close();
            return false;
        }
        for (VMIterator it = Q->visitMap.begin(); it != Q->visitMap.end(); it++){
             ActionT bestaction=getBestAction(it->first);
             for (AVIterator a = it->second.begin(); a != it->second.end();  ){
                  if (!(bestaction==a->first)) Q->delStateActionPair(it->first,(a++)->first);
                  else ++a;
            }
        }
        return saveTable(filename, isAppendMode);
    }

	bool readTable(const char* filename){

		std::ifstream in(filename);

		if(!in) {
			std::cout << "Cannot open input file.\n";
			return false;
		}
        Q->clearAll();

		int   _episode      = 0;
        float _alpha        = 0.0f;
		float _gamma        = 0.0f;
		float _rho          = 0.0f;
		bool  _preferTryNew = true;

        bool success = getAdditionalInfo(in);
        if (!success){
            std::cout<<"[ERROR] QLeaner: Cannot read Additional Information!"<<std::endl;
            exit(0);
        }

        in.ignore(20, ':');         in >> _episode;
        in.ignore(20, ':'); 		in >> _alpha;
        in.ignore(20, ':'); 		in >> _gamma;
        in.ignore(20, ':'); 		in >> _rho;
        in.ignore(20, ':'); 		in >> _preferTryNew;
		in >> *Q;
		in.close();

		if (!Q->isReadSuccess()){
            std::cout<<" CANNOT read the table " <<filename<<std::endl;
			Q->clearAll();
            return false;
		}

		episode = _episode;
		alpha   = _alpha;
		gamma   = _gamma;
		rho     = _rho;
		preferTryNew = _preferTryNew;
		Q->prepared = true;
		return true;
	}
};
#endif//__Q_LEARNING_H__
